##########################################
# DEPENDENCIES                           #
##########################################

from typing import Any as _Any

##########################################
# CONTEXT                                #
##########################################

class Context():
    def __init__(self, display, symbols=None, parent=None, parententry=None, caughterrors=None):
        self.display      = display
        self.parent       = parent
        self.parententry  = parententry

        self.symbols      = symbols

        self.caughterrors = []

    def caughterror(self, error):
        if self.parent:
            self.parent.caughterror(error)
        self.caughterrors.append(error)

    def copy(self):
        ctx = Context(self.display, self.symbols.copy(), self.parent, self.parententry, self.caughterrors)
        return(ctx)

##########################################
# SYMBOL TABLE                           #
##########################################

class SymbolTable():
    def __init__(self, parent=None):
        self.symbols = {}

        self.parent = parent


    def access(self, name: str) -> _Any:
        value = self.symbols.get(name, None)

        if value == None and self.parent:
            return(
                self.parent.access(name)
            )

        return(value)


    def assign(self, name: str, value: _Any):
        self.symbols[name] = value


    def remove(self, name: str):
        del self.symbols[name]

    def copy(self):
        sym = SymbolTable(self.parent)
        sym.symbols = self.symbols.copy()
        return(sym)


##########################################
# DEPENDENCIES                           #
##########################################

def _parserinit(conf, tokens, exceptions, constants, nodes):
    global lang
    lang = conf
    global TT_EOF, TT_EOL, TT_KEYWORD, TT_LPAREN, TT_RPAREN, TT_IDENTIFIER, TT_EQUALS, TT_EQEQUALS, TT_BANGEQUALS, TT_LESSTHAN, TT_LTEQUALS, TT_GREATERTHAN, TT_GTEQUALS, TT_PLUS, TT_DASH, TT_ASTRISK, TT_FSLASH, TT_CARAT, TT_LSQUARE, TT_RSQUARE, TT_PERIOD, TT_COMMA, TT_INT, TT_FLOAT, TT_STRING, TT_LCURLY, TT_RCURLY, TT_COLON, TT_ARROW
    TT_EOF          = tokens.TT_EOF
    TT_EOL          = tokens.TT_EOL
    TT_KEYWORD      = tokens.TT_KEYWORD
    TT_LPAREN       = tokens.TT_LPAREN
    TT_RPAREN       = tokens.TT_RPAREN
    TT_IDENTIFIER   = tokens.TT_IDENTIFIER
    TT_EQUALS       = tokens.TT_EQUALS
    TT_EQEQUALS     = tokens.TT_EQEQUALS
    TT_BANGEQUALS   = tokens.TT_BANGEQUALS
    TT_LESSTHAN     = tokens.TT_LESSTHAN
    TT_LTEQUALS     = tokens.TT_LTEQUALS
    TT_GREATERTHAN  = tokens.TT_GREATERTHAN
    TT_GTEQUALS     = tokens.TT_GTEQUALS
    TT_PLUS         = tokens.TT_PLUS
    TT_DASH         = tokens.TT_DASH
    TT_ASTRISK      = tokens.TT_ASTRISK
    TT_FSLASH       = tokens.TT_FSLASH
    TT_CARAT        = tokens.TT_CARAT
    TT_LSQUARE      = tokens.TT_LSQUARE
    TT_RSQUARE      = tokens.TT_RSQUARE
    TT_PERIOD       = tokens.TT_PERIOD
    TT_COMMA        = tokens.TT_COMMA
    TT_INT          = tokens.TT_INT
    TT_FLOAT        = tokens.TT_FLOAT
    TT_STRING       = tokens.TT_STRING
    TT_LCURLY       = tokens.TT_LCURLY
    TT_RCURLY       = tokens.TT_RCURLY
    TT_COLON        = tokens.TT_COLON
    TT_ARROW        = tokens.TT_ARROW
    global Syn_SyntaxError
    Syn_SyntaxError = exceptions.Syn_SyntaxError
    global KEYWORDS
    KEYWORDS        = constants.KEYWORDS
    global BinaryOpNode, ReturnNode, BreakNode, ContinueNode, IncludeNode, VarNullNode, VarCreateNode, UnaryOpNode, FuncCallNode, IndicieNode, AttributeNode, IntNode, FloatNode, StringNode, TupleNode, VarAssignNode, VarAccessNode, ArrayNode, DictionaryNode, FuncCreateNode, HandlerNode, IfNode, SwitchNode, ForLoopNode, WhileLoopNode
    BinaryOpNode    = nodes.BinaryOpNode
    ReturnNode      = nodes.ReturnNode
    BreakNode       = nodes.BreakNode
    ContinueNode    = nodes.ContinueNode
    IncludeNode     = nodes.IncludeNode
    VarNullNode     = nodes.VarNullNode
    VarCreateNode   = nodes.VarCreateNode
    UnaryOpNode     = nodes.UnaryOpNode
    FuncCallNode    = nodes.FuncCallNode
    IndicieNode     = nodes.IndicieNode
    AttributeNode   = nodes.AttributeNode
    IntNode         = nodes.IntNode
    FloatNode       = nodes.FloatNode
    StringNode      = nodes.StringNode
    TupleNode       = nodes.TupleNode
    VarAssignNode   = nodes.VarAssignNode
    VarAccessNode   = nodes.VarAccessNode
    ArrayNode       = nodes.ArrayNode
    DictionaryNode  = nodes.DictionaryNode
    FuncCreateNode  = nodes.FuncCreateNode
    HandlerNode     = nodes.HandlerNode
    IfNode          = nodes.IfNode
    SwitchNode      = nodes.SwitchNode
    ForLoopNode     = nodes.ForLoopNode
    WhileLoopNode   = nodes.WhileLoopNode

##########################################
# PARSE RESULT                           #
##########################################

class ParseResult():
    def __init__(self):
        self.error = None
        self.node = None
        self.advancecount = 0

    def registeradvancement(self):
        self.advancecount += 1

    def registerretreat(self):
        self.advancecount -= 1

    def register(self, res):
        self.advancecount += res.advancecount
        if res.error:
            self.error = res.error

        return(res.node)

    def tryregister(self, res):
        if res.error:
            return(None)

        return(res.node)

    def success(self, node):
        self.node = node

        return(self)

    def failure(self, error):
        if not self.error or self.advancecount == 0:
            self.error = error

        return(self)

##########################################
# PARSER                                 #
##########################################

class Parser():
    def __init__(self, tokens):
        self.tokens = tokens
        self.index = -1
        self.advance()

    def advance(self):
        self.index += 1
        if self.index < len(self.tokens):
            self.curtoken = self.tokens[self.index]

        return(self.curtoken)

    def retreat(self):
        self.index -= 1
        if self.index >= 0:
            self.curtoken = self.tokens[self.index]

        return(self.curtoken)


    def parse(self):
        res = ParseResult()
        tokens = []

        while self.curtoken.type != TT_EOF:
            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type == TT_EOF:
                break

            token = res.register(
                self.statement()
            )

            tokens.append(token)

            if res.error:
                return(
                    res
                )

            if self.curtoken.type != TT_EOL:
                msg = lang['exceptions']['syntaxerror']['invalideoflnoexp']
                msg = msg.replace('%s', 'L', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

        return(
            res.success(
                tokens
            )
        )


    def binaryop(self, funca, optypes, funcb=None):
        if not funcb:
            funcb = funca

        res = ParseResult()

        lfactor = res.register(funca())

        if res.error:
            return(res)

        while self.curtoken.type in optypes or (self.curtoken.type, self.curtoken.value) in optypes:
            optoken = self.curtoken
            res.registeradvancement()
            self.advance()
            rfactor = res.register(funcb())

            if res.error:
                return(res)

            lfactor = BinaryOpNode(
                lfactor, optoken, rfactor
            )

        return(
            res.success(lfactor)
        )


    def statement(self):
        res = ParseResult()
        start = self.curtoken.start.copy()

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['return']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            statement = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

            return(
                res.success(
                    ReturnNode(statement, start, end)
                )
            )

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['break']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

            return(
                res.success(
                    BreakNode(start, end)
                )
            )

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['continue']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

            return(
                res.success(
                    ContinueNode(start, end)
                )
            )

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['import']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            file = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

            return(
                res.success(
                    IncludeNode(file, start, end)
                )
            )

        expr = res.register(
            self.expr()
        )

        if res.error:
            return(res)

        return(
            res.success(expr)
        )


    def expr(self):
        res = ParseResult()

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['varcreate']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_IDENTIFIER:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', 'identifier', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            token = self.curtoken
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_EQUALS:
                return(
                    res.success(
                        VarNullNode(
                            token
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            expr = res.register(self.statement())

            if res.error:
                return(res)

            return(
                res.success(
                    VarCreateNode(
                        token, expr
                    )
                )
            )

        node = res.register(
            self.binaryop(
                self.compexpr,
                ((TT_KEYWORD, KEYWORDS['logicaland']), (TT_KEYWORD, KEYWORDS['logicalor']))
            )
        )

        if res.error:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', 'identifier, keyword, operator, type', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        return(
            res.success(node)
        )


    def compexpr(self):
        res = ParseResult()

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['logicalnot']):
            optoken = self.curtoken
            res.registeradvancement()
            self.advance()

            node = res.register(self.compexpr())

            if res.error:
                return(res)

            return(
                res.success(
                    UnaryOpNode(
                        optoken, node
                    )
                )
            )

        node = res.register(
            self.binaryop(
                self.arithexpr,
                (TT_EQEQUALS, TT_BANGEQUALS, TT_LESSTHAN, TT_LTEQUALS, TT_GREATERTHAN, TT_GTEQUALS)
            )
        )

        if res.error:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', 'identifier, operator, keyword, type', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        return(
            res.success(
                node
            )
        )


    def arithexpr(self):
        return(
            self.binaryop(
                self.term,
                (TT_PLUS, TT_DASH)
            )
        )


    def term(self):
        return(
            self.binaryop(
                self.raised,
                (TT_ASTRISK, TT_FSLASH)
            )
        )


    def raised(self):
        return(
            self.binaryop(
                self.factor,
                tuple([TT_CARAT]),
                self.term
            )
        )


    def factor(self):
        res = ParseResult()
        token = self.curtoken

        if token.type in (TT_PLUS, TT_DASH):
            res.registeradvancement()
            self.advance()

            factor = res.register(self.factor())

            if res.error:
                return(res)

            return(
                res.success(
                    UnaryOpNode(
                        token, factor
                    )
                )
            )

        atom = res.register(
            self.atom()
        )

        if res.error:
            return(res)

        while self.curtoken.type in [TT_LPAREN, TT_LSQUARE, TT_PERIOD]:
            atom = res.register(
                self.calttricie(atom)
            )

            if res.error:
                return(res)

        return(
            res.success(atom)
        )


    def calttricie(self, prevnode):
        res = ParseResult()

        if self.curtoken.type == TT_LPAREN:
            calls = []

            while self.curtoken.type == TT_LPAREN:
                args = []
                opts = {}

                res.registeradvancement()
                self.advance()

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                if self.curtoken.type == TT_RPAREN:
                    end = self.curtoken.end.copy()
                    res.registeradvancement()
                    self.advance()
                else:
                    if self.curtoken.type == TT_IDENTIFIER:
                        token = self.curtoken
                        res.registeradvancement()
                        self.advance()

                        if self.curtoken.type == TT_EQUALS:
                            res.registeradvancement()
                            self.advance()

                            opts[token.value] = res.register(
                                self.statement()
                            )

                        else:
                            res.registerretreat()
                            self.retreat()

                            args.append(
                                res.register(
                                    self.statement()
                                )
                            )

                    else:
                        args.append(
                            res.register(
                                self.statement()
                            )
                        )

                    if res.error:
                        msg = lang['exceptions']['syntaxerror']['notfound']
                        msg = msg.replace('%s', '\')\', identifier, operator, keyword, type', 1)
                        return(
                            res.failure(
                                Syn_SyntaxError(
                                    msg,
                                    self.curtoken.start, self.curtoken.end
                                )
                            )
                        )


                    while self.curtoken.type == TT_COMMA:
                        res.registeradvancement()
                        self.advance()

                        while self.curtoken.type == TT_EOL:
                            res.registeradvancement()
                            self.advance()

                        if self.curtoken.type == TT_IDENTIFIER:
                            token = self.curtoken
                            res.registeradvancement()
                            self.advance()

                            if self.curtoken.type == TT_EQUALS:
                                res.registeradvancement()
                                self.advance()

                                opts[token.value] = res.register(
                                    self.statement()
                                )

                            else:
                                res.registerretreat()
                                self.retreat()

                                args.append(
                                    res.register(
                                        self.statement()
                                    )
                                )
                        else:
                            args.append(
                                res.register(
                                    self.statement()
                                )
                            )

                        if res.error:
                            return(res)

                    while self.curtoken.type == TT_EOL:
                        res.registeradvancement()
                        self.advance()

                    if self.curtoken.type != TT_RPAREN:
                        msg = lang['exceptions']['syntaxerror']['notfound']
                        msg = msg.replace('%s', '\',\', \')\'', 1)
                        return(
                            res.failure(
                                Syn_SyntaxError(
                                    msg,
                                    self.curtoken.start, self.curtoken.end
                                )
                            )
                        )

                    end = self.curtoken.end.copy()
                    res.registeradvancement()
                    self.advance()

                calls.append((args, opts, end))

            return(
                res.success(
                    FuncCallNode(
                        prevnode,
                        calls
                    )
                )
            )


        elif self.curtoken.type == TT_LSQUARE:
            indicies = []
            end = self.curtoken.end.copy()

            while self.curtoken.type == TT_LSQUARE:
                res.registeradvancement()
                self.advance()

                indicies.append(
                    res.register(
                        self.statement()
                    )
                )

                if res.error:
                    return(res)

                if not self.curtoken.type == TT_RSQUARE:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', '\']\'', 1)
                    return(
                        res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

                end = self.curtoken.end.copy()
                res.registeradvancement()
                self.advance()

            return(
                res.success(
                    IndicieNode(
                        prevnode, indicies, end
                    )
                )
            )


        elif self.curtoken.type == TT_PERIOD:
            attributes = []
            end = self.curtoken.end.copy()

            while self.curtoken.type == TT_PERIOD:
                res.registeradvancement()
                self.advance()

                if not self.curtoken.type == TT_IDENTIFIER:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', 'identifier', 1)
                    return(
                        res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

                attributes.append(
                    self.curtoken
                )

                end = self.curtoken.end.copy()
                res.registeradvancement()
                self.advance()

            return(
                res.success(
                    AttributeNode(
                        prevnode, attributes, end
                    )
                )
            )

        else:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'(\', \'[\', \'.\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )


    def atom(self):
        res = ParseResult()
        token = self.curtoken


        if token.type == TT_INT:
            res.registeradvancement()
            self.advance()

            return(
                res.success(
                    IntNode(token)
                )
            )

        elif token.type == TT_FLOAT:
            self.advance()

            return(
                res.success(
                    FloatNode(token)
                )
            )

        elif token.type == TT_STRING:
            self.advance()

            return(
                res.success(
                    StringNode(token)
                )
            )

        elif token.type == TT_LPAREN:
            start = self.curtoken.start.copy()
            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            expr = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            if self.curtoken.type == TT_COMMA:
                expr = [expr]

                while self.curtoken.type == TT_COMMA:
                    res.registeradvancement()
                    self.advance()

                    while self.curtoken.type == TT_EOL:
                        res.registeradvancement()
                        self.advance()

                    expr.append(
                        res.register(
                            self.statement()
                        )
                    )
                
                if self.curtoken.type != TT_RPAREN:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', '\',\', \')\'', 1)
                    return(
                    res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

                end = self.curtoken.end.copy()
                res.registeradvancement()
                self.advance()

                return(
                    res.success(
                        TupleNode(
                            expr,
                            start, end
                        )
                    )
                )

            else:
                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                if self.curtoken.type == TT_RPAREN:
                    res.registeradvancement()
                    self.advance()

                    return(
                        res.success(
                            expr
                        )
                    )

            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\')\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )


        elif token.type == TT_IDENTIFIER:
            res.registeradvancement()
            self.advance()

            if self.curtoken.type == TT_EQUALS:
                res.registeradvancement()
                self.advance()

                expr = res.register(
                    self.statement()
                )

                return(
                    res.success(
                        VarAssignNode(
                            token,
                            expr
                        )
                    )
                )

            return(
                res.success(
                    VarAccessNode(token)
                )
            )

        elif token.type == TT_LSQUARE:
            arrayexpr = res.register(
                self.arrayexpr()
            )

            if res.error:
                return(res)

            return(
                res.success(
                    arrayexpr
                )
            )

        elif token.type == TT_LCURLY:
            dictionaryexpr = res.register(
                self.dictionaryexpr()
            )

            if res.error:
                return(res)

            return(
                res.success(
                    dictionaryexpr
                )
            )

        elif token.matches(TT_KEYWORD, KEYWORDS['funccreate']) or token.matches(TT_KEYWORD, KEYWORDS['lambdacreate']):
            funcdef = res.register(self.funcdef())
            if res.error:
                return(res)

            return(
                res.success(
                    funcdef
                )
            )

        elif token.matches(TT_KEYWORD, KEYWORDS['handler']):
            funcdef = res.register(self.handler())
            if res.error:
                return(res)

            return(
                res.success(
                    funcdef
                )
            )

        elif token.matches(TT_KEYWORD, KEYWORDS['if']):
            ifexpr = res.register(self.ifexpr())
            if res.error:
                return(res)

            return(
                res.success(
                    ifexpr
                )
            )

        elif token.matches(TT_KEYWORD, KEYWORDS['case']):
            switchexpr = res.register(self.switchexpr())
            if res.error:
                return(res)

            return(
                res.success(
                    switchexpr
                )
            )

        elif token.matches(TT_KEYWORD, KEYWORDS['forloop']) or token.matches(TT_KEYWORD, KEYWORDS['whileloop']):
            loopexpr = res.register(self.loopexpr())
            if res.error:
                return(res)

            return(
                res.success(
                    loopexpr
                )
            )

        msg = lang['exceptions']['syntaxerror']['notfound']
        msg = msg.replace('%s', 'identifier, keyword, operator, type', 1)
        return(
            res.failure(
                Syn_SyntaxError(
                    msg,
                    self.curtoken.start, self.curtoken.end
                )
            )
        )


    def arrayexpr(self):
        res = ParseResult()
        elmnodes = []
        start = self.curtoken.start.copy()

        if self.curtoken.type != TT_LSQUARE:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'[\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type == TT_RSQUARE:
            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

        else:
            elmnodes.append(
                res.register(
                    self.statement()
                )
            )

            if res.error:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\']\', identifier, keyword, operator, type', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            while self.curtoken.type == TT_COMMA:
                res.registeradvancement()
                self.advance()

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                elmnodes.append(
                    res.register(
                        self.statement()
                    )
                )

                if res.error:
                    return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RSQUARE:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\',\', \']\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

        return(
            res.success(
                ArrayNode(
                    elmnodes,
                    start, end
                )
            )
        )


    def dictionaryexpr(self):
        res = ParseResult()
        keys = []
        values = []
        start = self.curtoken.start.copy()

        if self.curtoken.type != TT_LCURLY:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'{\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type == TT_RCURLY:
            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()
        
        else:
            keys.append(
                res.register(
                    self.statement()
                )
            )

            if res.error:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'}\', identifier, keyword, operator, type', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            if self.curtoken.type != TT_COLON:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\':\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            values.append(
                res.register(
                    self.statement()
                )
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            while self.curtoken.type == TT_COMMA:
                res.registeradvancement()
                self.advance()

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                keys.append(
                    res.register(
                        self.statement()
                    )
                )

                if res.error:
                    return(res)

                if self.curtoken.type != TT_COLON:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', '\':\'', 1)
                    return(
                        res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

                res.registeradvancement()
                self.advance()

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                values.append(
                    res.register(
                        self.statement()
                    )
                )

                if res.error:
                    return(res)

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

            if self.curtoken.type != TT_RCURLY:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\',\', \'}\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken.end.copy()
            res.registeradvancement()
            self.advance()

        return(
            res.success(
                DictionaryNode(
                    keys, values,
                    start, end
                )
            )
        )

            


    def funcdef(self):
        res = ParseResult()

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['funccreate']) and not self.curtoken.matches(TT_KEYWORD, KEYWORDS['lambdacreate']):
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["funccreate"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        token = self.curtoken
        mode = self.curtoken.value

        res.registeradvancement()
        self.advance()

        if self.curtoken.type != TT_LPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'(\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        arguments = {}
        options   = {}

        if self.curtoken.type == TT_IDENTIFIER:
            argname = self.curtoken

            res.registeradvancement()
            self.advance()


            if self.curtoken.type == TT_COLON:
                res.registeradvancement()
                self.advance()

                argtype = res.register(
                    self.statement()
                )

                if res.error:
                    return(res)

                arguments[argname.value] = argtype

            elif self.curtoken.type == TT_EQUALS:
                res.registeradvancement()
                self.advance()

                optdefault = res.register(
                    self.statement()
                )

                if res.error:
                    return(res)

                options[argname.value] = optdefault

            else:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\':\', \'=\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            while self.curtoken.type == TT_COMMA:
                res.registeradvancement()
                self.advance()

                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                if self.curtoken.type != TT_IDENTIFIER:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', 'identifier', 1)
                    return(
                        res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

                argname = self.curtoken

                res.registeradvancement()
                self.advance()

                if self.curtoken.type == TT_COLON:
                    res.registeradvancement()
                    self.advance()

                    argtype = res.register(
                        self.statement()
                    )

                    if res.error:
                        return(res)

                    arguments[argname.value] = argtype

                elif self.curtoken.type == TT_EQUALS:
                    res.registeradvancement()
                    self.advance()

                    optdefault = res.register(
                        self.statement()
                    )

                    if res.error:
                        return(res)

                    options[argname.value] = optdefault

                else:
                    msg = lang['exceptions']['syntaxerror']['notfound']
                    msg = msg.replace('%s', '\':\', \'=\'', 1)
                    return(
                        res.failure(
                            Syn_SyntaxError(
                                msg,
                                self.curtoken.start, self.curtoken.end
                            )
                        )
                    )

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\',\', \')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

        elif self.curtoken.type != TT_RPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\')\', identifier', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_ARROW:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'->\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        returntype = res.register(
            self.statement()
        )

        if res.error:
            return(res)

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if mode == KEYWORDS['funccreate']:
            codeblock = res.register(
                self.codeblock()
            )

            if res.error:
                return(res)
        else:
            if self.curtoken.type != TT_LCURLY:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\', identifier', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
                
            line = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RCURLY:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'}\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            end = self.curtoken
            res.registeradvancement()
            self.advance()

            codeblock = [ReturnNode(
                line, token.start, end
            )]

        return(
            res.success(
                FuncCreateNode(
                    token,
                    arguments,
                    options,
                    returntype,
                    codeblock,
                    False
                )
            )
        )


    def handler(self):
        res = ParseResult()

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['handler']):
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["handler"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        token = self.curtoken

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        codeblock = res.register(
            self.codeblock()
        )

        if res.error:
            return(res)

        return(
            res.success(
                HandlerNode(
                    token,
                    codeblock
                )
            )
        )


    def ifexpr(self):
        res = ParseResult()
        cases = []
        elsecase = None

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['if']):
            
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["if"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        if self.curtoken.type != TT_LPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'(\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )
        
        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        condition = res.register(self.statement())
        if res.error: return(res)

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_RPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\')\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )
        
        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        codeblock = res.register(self.codeblock())

        if res.error:
            return(res)

        cases.append((condition, codeblock))

        while self.curtoken.matches(TT_KEYWORD, KEYWORDS['elif']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )
        
            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            condition = res.register(self.statement())
            if res.error: return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )
        
            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            codeblock = res.register(self.codeblock())

            if res.error:
                return(res)
            cases.append((condition, codeblock))

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['else']):
            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
            
            elsecase = res.register(self.codeblock())
            if res.error:
                return(res)

        return(
            res.success(
                IfNode(
                    cases, elsecase
                )
            )
        )


    def switchexpr(self):
        res = ParseResult()
        cases = []
        elsecase = None

        start = self.curtoken.start

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['case']):
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["case"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        if self.curtoken.type != TT_LPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'(\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )
        
        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        varoverwrite = False
        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['varcreate']):
            varoverwrite = True

            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_IDENTIFIER:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', 'identifier', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        vartoken = self.curtoken

        res.registeradvancement()
        self.advance()

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['as']):
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["as"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        value = res.register(
            self.statement()
        )

        if res.error:
            return(res)

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_RPAREN:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\')\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_LCURLY:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'{\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['when']):
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', f'{KEYWORDS["when"]}', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        cases = []
        while self.curtoken.matches(TT_KEYWORD, KEYWORDS['when']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            condition = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            codeblock = res.register(
                self.codeblock()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            cases.append((condition, codeblock))

        elsecase = None
        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['else']):
            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
            
            elsecase = res.register(self.codeblock())
            if res.error:
                return(res)

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_RCURLY:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'}\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        return(
            res.success(
                SwitchNode(
                    vartoken, varoverwrite, value,
                    cases, elsecase,
                    start, self.curtoken.end
                )
            )
        )



    def loopexpr(self):
        res = ParseResult()

        if self.curtoken.matches(TT_KEYWORD, KEYWORDS['forloop']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            varoverwrite = False
            if self.curtoken.matches(TT_KEYWORD, KEYWORDS['varcreate']):
                varoverwrite = True

                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_IDENTIFIER:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', 'identifier', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            vartoken = self.curtoken

            res.registeradvancement()
            self.advance()

            if not self.curtoken.matches(TT_KEYWORD, KEYWORDS['in']):
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', f'{KEYWORDS["in"]}', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            loopthrough = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
            
            codeblock = res.register(
                self.codeblock()
            )

            if res.error:
                return(res)

            return(
                res.success(
                    ForLoopNode(
                        vartoken, varoverwrite, loopthrough, codeblock
                    )
                )
            )


        elif self.curtoken.matches(TT_KEYWORD, KEYWORDS['whileloop']):
            res.registeradvancement()
            self.advance()

            if self.curtoken.type != TT_LPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\'(\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
            
            condition = res.register(
                self.statement()
            )

            if res.error:
                return(res)

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()
            
            if self.curtoken.type != TT_RPAREN:
                msg = lang['exceptions']['syntaxerror']['notfound']
                msg = msg.replace('%s', '\')\'', 1)
                return(
                    res.failure(
                        Syn_SyntaxError(
                            msg,
                            self.curtoken.start, self.curtoken.end
                        )
                    )
                )

            res.registeradvancement()
            self.advance()

            while self.curtoken.type == TT_EOL:
                res.registeradvancement()
                self.advance()

            codeblock = res.register(
                self.codeblock()
            )

            if res.error:
                return(res)

            return(
                res.success(
                    WhileLoopNode(
                        condition, codeblock
                    )
                )
            )


        else:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'{KEYWORDS["forloop"]}\', \'KEYWORDS["whileloop"]\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )



    def codeblock(self):
        res = ParseResult()
        tokens = []

        if self.curtoken.type != TT_LCURLY:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'{\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        while self.curtoken.type == TT_EOL:
            res.registeradvancement()
            self.advance()

        if self.curtoken.type != TT_RCURLY:
            expr = res.register(
                self.statement()
            )

            if res.error:
                return(res)
            tokens.append(expr)

            while self.curtoken.type == TT_EOL:
                while self.curtoken.type == TT_EOL:
                    res.registeradvancement()
                    self.advance()

                if self.curtoken.type == TT_RCURLY:
                    break

                expr = res.register(
                    self.statement()
                )

                if res.error:
                    return(res)
                tokens.append(expr)

        if self.curtoken.type != TT_RCURLY:
            msg = lang['exceptions']['syntaxerror']['notfound']
            msg = msg.replace('%s', '\'}\'', 1)
            return(
                res.failure(
                    Syn_SyntaxError(
                        msg,
                        self.curtoken.start, self.curtoken.end
                    )
                )
            )

        res.registeradvancement()
        self.advance()

        return(
            res.success(tokens)
        )



##########################################
# DEPENDENCIES                           #
##########################################

from pathlib import Path as _Path
from importlib import import_module as _import_module
import re as _re

def _interpreterinit(conf, tokens, context, default, constants, types, exceptions, runu, perimodu):
    global lang
    lang = conf
    global TT_ASTRISK, TT_BANGEQUALS, TT_CARAT, TT_DASH, TT_EQEQUALS, TT_FSLASH, TT_GREATERTHAN, TT_GTEQUALS, TT_KEYWORD, TT_LESSTHAN, TT_LTEQUALS, TT_PLUS
    TT_ASTRISK          = tokens.TT_ASTRISK
    TT_BANGEQUALS       = tokens.TT_BANGEQUALS
    TT_CARAT            = tokens.TT_CARAT
    TT_DASH             = tokens.TT_DASH
    TT_EQEQUALS         = tokens.TT_EQEQUALS
    TT_FSLASH           = tokens.TT_FSLASH
    TT_GREATERTHAN      = tokens.TT_GREATERTHAN
    TT_GTEQUALS         = tokens.TT_GTEQUALS
    TT_KEYWORD          = tokens.TT_KEYWORD
    TT_LESSTHAN         = tokens.TT_LESSTHAN
    TT_LTEQUALS         = tokens.TT_LTEQUALS
    TT_PLUS             = tokens.TT_PLUS
    global Context, SymbolTable
    Context             = context.Context
    SymbolTable         = context.SymbolTable
    global defaultvariables
    defaultvariables    = default.defaultvariables
    global KEYWORDS, RESERVED
    KEYWORDS            = constants.KEYWORDS
    RESERVED            = constants.RESERVED
    global TYPES, NullType, IntType, FloatType, StringType, ArrayType, TupleType, DictionaryType, FunctionType, BuiltInFunctionType, ExceptionType, NamespaceType
    TYPES               = types.TYPES
    NullType            = types.NullType
    IntType             = types.IntType
    FloatType           = types.FloatType
    StringType          = types.StringType
    ArrayType           = types.ArrayType
    TupleType           = types.TupleType
    DictionaryType      = types.DictionaryType
    FunctionType        = types.FunctionType
    BuiltInFunctionType = types.BuiltInFunctionType
    ExceptionType       = types.ExceptionType
    NamespaceType       = types.NamespaceType
    global Exc_BreakError, Exc_ContinueError, Exc_FileAccessError, Exc_IdentifierError, Exc_IncludeError, Exc_IterationError, Exc_PanicError, Exc_PatternError, Exc_ReservedError, Exc_ReturnError, Exc_TypeError
    Exc_BreakError      = exceptions.Exc_BreakError
    Exc_ContinueError   = exceptions.Exc_ContinueError
    Exc_FileAccessError = exceptions.Exc_FileAccessError
    Exc_IdentifierError = exceptions.Exc_IdentifierError
    Exc_IncludeError    = exceptions.Exc_IncludeError
    Exc_IterationError  = exceptions.Exc_IterationError
    Exc_PatternError    = exceptions.Exc_PatternError
    Exc_PanicError      = exceptions.Exc_PanicError
    Exc_ReservedError   = exceptions.Exc_ReservedError
    Exc_ReturnError     = exceptions.Exc_ReturnError
    Exc_TypeError       = exceptions.Exc_TypeError
    global run
    run = runu.run
    global perimod
    perimod = perimodu

##########################################
# RUNTIME RESULT                         #
##########################################

class RTResult():
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = None
        self.funcvalue = None
        self.shouldbreak = False
        self.shouldcontinue = False
        self.error = None

    def register(self, res):
        self.error = res.error
        self.funcvalue = res.funcvalue
        self.shouldbreak = res.shouldbreak
        self.shouldcontinue = res.shouldcontinue

        return(res.value)

    def success(self, value):
        self.reset()
        self.value = value

        return(self)

    def successreturn(self, value):
        self.reset()
        self.funcvalue = value

        return(self)

    def successbreak(self, value):
        self.reset()
        self.shouldbreak = value

        return(self)

    def successcontinue(self, value):
        self.reset()
        self.shouldcontinue = value

        return(self)

    def failure(self, error):
        self.error = error

        return(self)

    def shouldreturn(self):
        return(
            self.error or self.funcvalue or self.shouldbreak or self.shouldcontinue
        )

##########################################
# INTERPRETER                            #
##########################################

class Interpreter():
    def visit(self, node, context, insideloop=False):
        method = f'visit_{type(node).__name__}'
        meth = method
        method = getattr(self, method)

        result = method(node, context, insideloop=insideloop)

        return(result)



    ### TYPES
    def visit_IntNode(self, node, context, insideloop=False):
        return(
            RTResult().success(
                IntType(node.token.value)
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )


    def visit_FloatNode(self, node, context, insideloop=False):
        return(
            RTResult().success(
                FloatType(node.token.value)
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )


    def visit_StringNode(self, node, context, insideloop=False):
        return(
            RTResult().success(
                StringType(node.token.value)
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )


    def visit_ArrayNode(self, node, context, insideloop=False):
        res = RTResult()
        elements = []
        type_ = None
        for i in node.elmnodes:
            elm = res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            if type_:
                if type(elm) != type(type_):
                    msg = lang['exceptions']['typeerror']['cannot']
                    msg = msg.replace('%s', TYPES['list'], 1)
                    msg = msg.replace('%s', type_.type, 1)
                    msg = msg.replace('%s', 'include', 1)
                    msg = msg.replace('%s', elm.type, 1)
                    return(
                        res.failure(
                            Exc_TypeError(
                                msg,
                                elm.start, elm.end,
                                context,
                                elm.originstart, elm.originend, elm.origindisplay
                            )
                        )
                    )
            else:
                type_ = elm

            elements.append(elm)

        return(
            res.success(
                ArrayType(elements)
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )


    def visit_TupleNode(self, node, context, insideloop=False):
        res = RTResult()
        elements = []
        for i in node.elmnodes:
            elm = res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            elements.append(elm)

        return(
            res.success(
                TupleType(tuple(elements))
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )


    def visit_DictionaryNode(self, node, context, insideloop=False):
        res = RTResult()
        elements = {}
        keytype = None
        valuetype = None
        for i in range(len(node.keynodes)):
            key = res.register(
                self.visit(
                    node.keynodes[i],
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            value = res.register(
                self.visit(
                    node.valuenodes[i],
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            if keytype:
                if type(key) != type(keytype):
                    msg = lang['exceptions']['typeerror']['cannot']
                    msg = msg.replace('%s', TYPES['dictionary'] + ' keys', 1)
                    msg = msg.replace('%s', keytype.type, 1)
                    msg = msg.replace('%s', 'include', 1)
                    msg = msg.replace('%s', key.type, 1)
                    return(
                        res.failure(
                            Exc_TypeError(
                                msg,
                                key.start, key.end,
                                context,
                                key.originstart, key.originend, key.origindisplay
                            )
                        )
                    )
            else:
                keytype = key

            if valuetype:
                msg = lang['exceptions']['typeerror']['cannot']
                msg = msg.replace('%s', TYPES['dictionary'], 1)
                msg = msg.replace('%s', valuetype.type, 1)
                msg = msg.replace('%s', 'include', 1)
                msg = msg.replace('%s', value.type, 1)
                if type(value) != type(valuetype):
                    return(
                        res.failure(
                            Exc_TypeError(
                                msg,
                                value.start, value.end,
                                context,
                                value.originstart, value.originend, value.origindisplay
                            )
                        )
                    )
            else:
                valuetype = value

            if res.shouldreturn():
                return(res)

            elements[key] = value

        return(
            res.success(
                DictionaryType(elements)
                    .setcontext(context)
                    .setpos(node.start, node.end)
            )
        )



    ### VARIABLE CONTROL
    def visit_VarAccessNode(self, node, context, insideloop=False):
        res = RTResult()

        name = node.token.value
        value = context.symbols.access(name)

        if not value:
            msg = lang['exceptions']['identifiererror']['notdefined']
            msg = msg.replace('%s', name, 1)
            return(
                res.failure(
                    Exc_IdentifierError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        value = value.copy().setcontext(context)

        value.originstart.append(value.start)
        value.originend.append(value.end)
        value.origindisplay.append(context.display)

        value.start = node.start
        value.end = node.end

        value.name = name

        return(
            res.success(
                value
            )
        )


    def visit_VarAssignNode(self, node, context, insideloop=False):
        res = RTResult()

        name = node.token.value

        value = res.register(
            self.visit(
                node.valnode,
                context,
                    insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        prevvalue = context.symbols.access(name)

        if not prevvalue:
            msg = lang['exceptions']['identifiererror']['notdefined']
            msg = msg.replace('%s', name, 1)
            return(
                res.failure(
                    Exc_IdentifierError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        if name in RESERVED or prevvalue.reserved:
            msg = lang['exceptions']['reservederror']['reserved']
            msg = msg.replace('%s', str(value), 1)
            msg = msg.replace('%s', name, 1)
            return(
                res.failure(
                    Exc_ReservedError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        if type(prevvalue) != type(value) and not isinstance(prevvalue, NullType):
            value.originstart += prevvalue.originstart
            value.originend += prevvalue.originend
            value.origindisplay += prevvalue.origindisplay
            msg = lang['exceptions']['typeerror']['assigntype']
            msg = msg.replace('%s', value.type, 1)
            msg = msg.replace('%s', name, 1)
            msg = msg.replace('%s', prevvalue.type, 1)
            return(
                res.failure(
                    Exc_TypeError(
                        msg,
                        node.valnode.start, node.valnode.end,
                        context,
                        value.originstart, value.originend, value.origindisplay
                    )
                )
            )

        if res.shouldreturn():
            return(res)

        value.name = name

        context.symbols.assign(name, value)

        return(
            res.success(
                value
            )
        )


    def visit_VarCreateNode(self, node, context, insideloop=False):
        res = RTResult()

        name = node.token.value
        value = res.register(
            self.visit(
                node.valnode,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        prevvalue = context.symbols.access(name)

        if prevvalue:
            if name in RESERVED or prevvalue.reserved:
                msg = lang['exceptions']['reservederror']['reserved']
                msg = msg.replace('%s', str(value), 1)
                msg = msg.replace('%s', name, 1)
                return(
                    res.failure(
                        Exc_ReservedError(
                            msg,
                            node.start, node.end,
                            context
                        )
                    )
                )

        value.name = name

        context.symbols.assign(name, value)

        return(
            res.success(
                value
            )
        )


    def visit_VarNullNode(self, node, context, insideloop=False):
        res = RTResult()

        name = node.token

        prevvalue = context.symbols.access(name)

        if prevvalue:
            if name in RESERVED or prevvalue.reserved:
                msg = lang['exceptions']['reservederror']['reserved']
                msg = msg.replace('%s', TYPES['nonetype'], 1)
                msg = msg.replace('%s', name, 1)
                return(
                    res.failure(
                        Exc_ReservedError(
                            msg,
                            node.start, node.end,
                            context
                        )
                    )
                )

        value = NullType().setpos(node.start, node.end).setcontext(context)

        value.name = name

        context.symbols.assign(
            name.value,
            value
        )

        return(
            res.success(value)
        )



    ### FUNCTIONS
    def visit_FuncCreateNode(self, node, context, insideloop=False):
        res = RTResult()

        bodynodes = node.bodynodes
        arguments = node.arguments
        options = node.options
        returntype = node.returntype

        for key in list(arguments.keys()):
            i = arguments[key]
            
            i = res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            if i.type != TYPES['type'] and not isinstance(i, NullType):
                msg = lang['exceptions']['typeerror']['mustbe']
                msg = msg.replace('%s', 'Argument type', 1)
                msg = msg.replace('%s', f'{TYPES["type"]} or {TYPES["nonetype"]}', 1)
                msg = msg.replace('%s', i.type, 1)
                return(
                    res.failure(
                        Exc_TypeError(
                            msg,
                            i.start, i.end,
                            context
                        )
                    )
                )

            if isinstance(i, NullType):
                arguments[key] = NullType
            else:
                arguments[key] = i.returntype

        for key in list(options.keys()):
            i = options[key]

            i = res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            options[key] = i

        returntype = res.register(
            self.visit(
                returntype,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        if returntype.type != TYPES['type'] and not isinstance(returntype, NullType):
            msg = lang['exceptions']['typeerror']['mustbe']
            msg = msg.replace('%s', 'Return type', 1)
            msg = msg.replace('%s', f'{TYPES["type"]} or {TYPES["nonetype"]}', 1)
            msg = msg.replace('%s', returntype.type, 1)
            return(
                res.failure(
                    Exc_TypeError(
                        msg,
                        returntype.start, returntype.end,
                        context
                    )
                )
            )

        if isinstance(returntype, NullType):
            funcvalue = FunctionType(bodynodes, arguments, options, NullType, node.shouldreturn)
        else:
            funcvalue = FunctionType(bodynodes, arguments, options, returntype, node.shouldreturn)
        funcvalue.setcontext(context).setpos(node.start, node.end)

        return(
            res.success(funcvalue)
        )


    def visit_FuncCallNode(self, node, context, insideloop=False):
        res = RTResult()

        value = res.register(
            self.visit(
                node.node,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        for i in node.calls:
            args = []
            for argnode in i[0]:
                args.append(
                    res.register(
                        self.visit(
                            argnode,
                            context,
                            insideloop=insideloop
                        )
                    )
                )
            opts = {}
            for optnode in list(i[1].keys()):
                opts[optnode] = res.register(
                    self.visit(
                        i[1][optnode],
                        context,
                        insideloop=insideloop
                    )
                )

            if res.shouldreturn():
                return(res)

            try:
                callvalue = value.name
            except AttributeError:
                callvalue = None

            value = res.register(
                value.call(
                    callvalue, args, opts, i[0]
                )
            )

            if res.shouldreturn():
                return(res)

            value = value.copy().setpos(node.start, node.end).setcontext(context)

        return(
            res.success(value)
        )


    def visit_ReturnNode(self, node, context, insideloop=False):
        res = RTResult()

        if not context.parent:
            msg = lang['exceptions']['returnerror']['location']
            return(
                res.failure(
                    Exc_ReturnError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        if node.returnnode:
            value = res.register(
                self.visit(
                    node.returnnode,
                    context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)
        else:
            value = NullType()

        return(
            res.successreturn(value)
        )
    


    ### FLOW CONTROL
    def visit_IfNode(self, node, context, insideloop=False):
        res = RTResult()
        #returnval = NullType().setpos(node.start, node.end).setcontext(context)

        for condition, codeblock in node.cases:
            condvalue = res.register(
                self.visit(
                    condition,
                    context,
                    insideloop=insideloop
                )
            )
            if res.shouldreturn():
                return(res)

            istrue, error = condvalue.istrue()

            if error:
                return(
                    res.failure(
                        error
                    )
                )

            if istrue:
                for j in codeblock:
                    res.register(
                        self.visit(
                            j,
                            context,
                            insideloop=insideloop
                        )
                    )

                    if res.shouldreturn():
                        return(res)

                return(
                    res.success(
                        NullType()
                            .setpos(node.start, node.end)
                            .setcontext(context)
                    )
                )

        if node.elsecase:
            for i in node.elsecase:
                res.register(
                    self.visit(
                        i,
                        context,
                        insideloop=insideloop
                    )
                )

                if res.shouldreturn():
                    return(res)
            
        return(
            res.success(
                NullType()
                    .setpos(node.start, node.end)
                    .setcontext(context)
            )
        )


    def visit_SwitchNode(self, node, context, insideloop=False):
        res = RTResult()

        varname = node.vartoken.value
        value = res.register(
            self.visit(
                node.value,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        prevvalue = context.symbols.access(varname)
        if not node.varoverwrite and not prevvalue:
            msg = lang['exceptions']['identifiererror']['notdefined']
            msg = msg.replace('%s', varname, 1)
            return(
                res.failure(
                    Exc_IdentifierError(
                        msg,
                        node.vartoken.start, node.vartoken.end,
                        context
                    )
                )
            )
        if type(value) != type(prevvalue) and not node.varoverwrite:
            msg = lang['exceptions']['typeerror']['assigntype']
            msg = msg.replace('%s', value.type, 1)
            msg = msg.replace('%s', varname, 1)
            msg = msg.replace('%s', prevvalue.type, 1)
            return(
                res.failure(
                    Exc_TypeError(
                        msg,
                        value.start, value.end,
                        context,
                        value.originstart, value.originend, value.origindisplay
                    )
                )
            )

        value.reserved = True

        exec_symbols = SymbolTable(context.symbols.parent)
        exec_symbols.assign(varname, value)
        exec_context = Context(context.display, exec_symbols, context.parent, context.parententry)

        context.symbols.assign(varname, value)

        for condition, codeblock in node.cases:
            condvalue = res.register(
                self.visit(
                    condition,
                    exec_context,
                    insideloop=insideloop
                )
            )

            if res.shouldreturn():
                return(res)

            istrue, error = condvalue.istrue()

            if error:
                return(
                    res.failure(
                        error
                    )
                )

            if istrue:
                for j in codeblock:
                    res.register(
                        self.visit(
                            j,
                            context,
                            insideloop=insideloop
                        )
                    )

                    if res.shouldreturn():
                        return(res)

                return(
                    res.success(
                        NullType()
                            .setpos(node.start, node.end)
                            .setcontext(context)
                    )
                )

        if node.elsecase:
            for i in node.elsecase:
                res.register(
                    self.visit(
                        i,
                        context,
                        insideloop=insideloop
                    )
                )

                if res.shouldreturn():
                    return(res)

        return(
            res.success(
                NullType()
                    .setpos(node.start, node.end)
                    .setcontext(context)
            )
        )


    def visit_ForLoopNode(self, node, context, insideloop=False):
        res = RTResult()

        varname = node.vartoken.value

        loopthrough = res.register(
            self.visit(
                node.loopthrough,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        prevvalue = context.symbols.access(varname)
        if not node.varoverwrite and not prevvalue:
            msg = lang['exceptions']['identifiererror']['notdefined']
            msg = msg.replace('%s', varname, 1)
            return(
                res.failure(
                    Exc_IdentifierError(
                        msg,
                        node.vartoken.start, node.vartoken.end,
                        context
                    )
                )
            )

        try:
            shouldbreak = False
            for i in loopthrough.value:
                if type(i) != type(prevvalue) and not node.varoverwrite:
                    msg = lang['exceptions']['typeerror']['assigntype']
                    msg = msg.replace('%s', i.type, 1)
                    msg = msg.replace('%s', varname, 1)
                    msg = msg.replace('%s', prevvalue.type, 1)
                    return(
                        res.failure(
                            Exc_TypeError(
                                msg,
                                i.start, i.end,
                                context,
                                i.originstart, i.originend, i.origindisplay
                            )
                        )
                    )

                i.reserved = True
                context.symbols.assign(varname, i)

                for j in node.bodynodes:
                    res.register(
                        self.visit(
                            j,
                            context,
                            insideloop=True
                        )
                    )

                    if res.error:
                        return(res)

                    if res.shouldbreak:
                        shouldbreak = True
                        break
                
                    if res.shouldcontinue:
                        break

                if shouldbreak:
                    break

        except TypeError:
            msg = lang['exceptions']['iterationerror']['cannot']
            msg = msg.replace('%s', loopthrough.type, 1)
            return(
                res.failure(
                    Exc_IterationError(
                        msg,
                        loopthrough.start, loopthrough.end,
                        context
                    )
                )
            )

        return(
            res.success(
                NullType()
                    .setpos(node.start, node.end)
                    .setcontext(context)
            )
        )


    def visit_WhileLoopNode(self, node, context, insideloop=False):
        res = RTResult()

        while True:
            condition = res.register(
                self.visit(
                    node.condition,
                    context,
                    insideloop=insideloop
                )
            )

            if res.error:
                return(res)

            istrue, error = condition.istrue()

            if error:
                return(
                    res.failure(error)
                )

            if not istrue: break

            for j in node.bodynodes:
                res.register(
                    self.visit(
                        j,
                        context,
                        insideloop=insideloop
                    )
                )

                if res.shouldreturn():
                    return(res)

        return(
            res.success(
                NullType()
                    .setpos(node.start, node.end)
                    .setcontext(context)
            )
        )


    def visit_BreakNode(self, node, context, insideloop=False):
        res = RTResult()

        if not insideloop:
            msg = lang['exceptions']['breakerror']['location']
            return(
                res.failure(
                    Exc_BreakError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        res.successbreak(True)

        return(res)


    def visit_ContinueNode(self, node, context, insideloop=False):
        res = RTResult()

        if not insideloop:
            msg = lang['exceptions']['continueerror']['location']
            return(
                res.failure(
                    Exc_ContinueError(
                        msg,
                        node.start, node.end,
                        context
                    )
                )
            )

        res.successcontinue(True)

        return(res)



    ### HANDLER
    def visit_HandlerNode(self, node, context, insideloop=False):
        res = RTResult()

        for i in node.bodynodes:
            res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            if res.error:
                error = res.error
                if isinstance(error, Exc_PanicError):
                    return(res)

                context.caughterror(error)
                exc = ExceptionType(
                    error.exc,
                    error.msg,
                    error.start,
                    error.end,
                    context
                )

                try:
                    exc.setcontext(context).setpos(error.start, error.end, error.originstart, error.originend, error.origindisplay)

                except AttributeError:
                    return(res)

                return(
                    res.success(exc)
                )

            if res.shouldreturn():
                return(res)

        return(
            res.success(
                NullType()
                    .setpos(node.start, node.end)
                    .setcontext(context)
            )
        )



    ### OPERATIONS
    def visit_UnaryOpNode(self, node, context, insideloop=False):
        res = RTResult()
        result = res.register(
            self.visit(
                node.node,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        error = None

        if node.optoken.type == TT_DASH:
            result, error = result.multiply(IntType(-1))
        elif node.optoken.matches(TT_KEYWORD, KEYWORDS['logicalnot']):
            result, error = result.not_()

        if error:
            return(
                res.failure(
                    error
                )
            )

        return(
            res.success(
                result.setpos(
                    node.start,
                    node.end
                )
            )
        )


    def visit_BinaryOpNode(self, node, context, insideloop=False):
        res = RTResult()

        left = res.register(
            self.visit(
                node.lnode,
                context,
                insideloop=insideloop
            )
        )
        if res.shouldreturn():
            return(res)

        right = res.register(
            self.visit(
                node.rnode,
                context,
                insideloop=insideloop
            )
        )
        if res.shouldreturn():
            return(res)

        if node.optoken.type == TT_PLUS:
            result, error = left.add(right)
        elif node.optoken.type == TT_DASH:
            result, error = left.subtract(right)
        elif node.optoken.type == TT_ASTRISK:
            result, error = left.multiply(right)
        elif node.optoken.type == TT_FSLASH:
            result, error = left.divide(right)
        elif node.optoken.type == TT_CARAT:
            result, error = left.raised(right)
        elif node.optoken.type == TT_EQEQUALS:
            result, error = left.eqequals(right)
        elif node.optoken.type == TT_BANGEQUALS:
            result, error = left.bangequals(right)
        elif node.optoken.type == TT_LESSTHAN:
            result, error = left.lessthan(right)
        elif node.optoken.type == TT_GREATERTHAN:
            result, error = left.greaterthan(right)
        elif node.optoken.type == TT_LTEQUALS:
            result, error = left.ltequals(right)
        elif node.optoken.type == TT_GTEQUALS:
            result, error = left.gtequals(right)
        elif node.optoken.matches(TT_KEYWORD, KEYWORDS['logicaland']):
            result, error = left.and_(right)
        elif node.optoken.matches(TT_KEYWORD, KEYWORDS['logicalor']):
            result, error = left.or_(right)
        else:
            result = NullType()
            error = None

        if error:
            return(
                res.failure(error)
            )

        return(
            res.success(
                result.setpos(
                    node.start,
                    node.end
                ).setcontext(
                    context
                )
            )
        )



    ### MISCELLANIOUS
    def visit_IndicieNode(self, node, context, insideloop=False):
        res = RTResult()

        value = res.register(
            self.visit(
                node.node,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        for i in node.indicies:
            i = res.register(
                self.visit(
                    i,
                    context,
                    insideloop=insideloop
                )
            )

            value, error = value.indicie(i)

            if error:
                return(
                    res.failure(error)
                )

        return(
            res.success(value)
        )

    def visit_AttributeNode(self, node, context, insideloop=False):
        res = RTResult()

        value = res.register(
            self.visit(
                node.node,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        for i in node.attributes:
            value, error = value.attribute(i)

            if error:
                return(
                    res.failure(error)
                )

            if isinstance(value, FunctionType):
                value.name = i.value

        return(
            res.success(value)
        )



    ### MODULES
    def visit_IncludeNode(self, node, context, insideloop=False):
        res = RTResult()

        file = res.register(
            self.visit(
                node.filenode,
                context,
                insideloop=insideloop
            )
        )

        if res.shouldreturn():
            return(res)

        if not isinstance(file, StringType):
            msg = lang['exceptions']['typeerror']['mustbe']
            msg = msg.replace('%s', '\'file\'', 1)
            msg = msg.replace('%s', TYPES['string'], 1)
            msg = msg.replace('%s', file.type, 1)
            return(
                res.failure(
                    Exc_TypeError(
                        msg,
                        file.start, file.end,
                        context,
                        file.originstart, file.originend, file.origindisplay
                    )
                )
            )

        if not _re.match(r'^([A-z][A-z0-9]*)(\.([A-z][A-z0-9]*))*$', file.value):
            msg = lang['exceptions']['patternerror']['nomatch']
            msg = msg.replace('%s', '\'file\'', 1)
            msg = msg.replace('%s', '/^([A-z][A-z0-9]*)(\.([A-z][A-z0-9]*))*$/', 1)
            return(
                res.failure(
                    Exc_PatternError(
                        msg,
                        file.start, file.end,
                        context,
                        file.originstart, file.originend, file.origindisplay
                    )
                )
            )

        script = None
        for i in context.symbols.access('__peridot__').symbols.access('path').value:
            i = i.value

            try:
                path = file.value + '.peri'
                path = _Path(i) / path
                with open(str(path), 'r') as f:
                    script = f.read()
                break

            except FileNotFoundError as e: pass
            except IsADirectoryError as e: pass
            except PermissionError   as e: pass

        result = NullType()
        result.setcontext(
            context
        ).setpos(
            node.start, node.end
        )

        if not script:
            try:
                symbols = defaultvariables(SymbolTable(context.symbols))
                perimod._context = Context('<file>', symbols, context, [node.start, node.end, [file.originstart], [file.originend], [file.origindisplay]])
                perimod._symbols = symbols
                perimod._position = node
                perimod._file = file.value
                mod = _import_module(f'.version.modules.{file.value}', package='peridot')
                result = perimod._namespace
                if result == None:
                    msg = lang['exceptions']['includeerror']['failed']
                    msg = msg.replace('%s', str(file), 1)
                    msg = msg.replace('%s', lang['exceptions']['module'], 1)
                    msg = msg.replace('%s', perimod._error[0], 1)
                    msg = msg.replace('%s', lang['exceptions']['line'], 1)
                    msg = msg.replace('%s', str(perimod._error[1]), 1)
                    msg = msg.replace('%s', perimod._error[2], 1)
                    msg = msg.replace('%s', perimod._error[3], 1)
                    return(
                        res.failure(
                            Exc_IncludeError(
                                msg,
                                node.start, node.end,
                                context
                            )
                        )
                    )
                for i in list(perimod._builtinfuncs.keys()):
                    BuiltInFunctionType.modules[i] = perimod._builtinfuncs[i]
                result.setpos(node.start, node.end)
                result.setcontext(context)
                script = True

            except PermissionError:
                msg = lang['exceptions']['includeerror']['doesnotexist']
                msg = msg.replace('%s', str(file), 1)
                return(
                    res.failure(
                        Exc_FileAccessError(
                            msg,
                            file.start, file.end,
                            context,
                            file.originstart, file.originend, file.origindisplay
                        )
                    )
                )

        if not script:
            symbols = defaultvariables(SymbolTable())
            result, exec_context, error = run(file.value, script, symbols)

            if error:
                return(
                    res.failure(error)
                )

            result = NamespaceType(symbols).setpos(node.start, node.end).setcontext(exec_context)

        return(
            res.success(result)
        )



VERSION       = '1.0.0'
MODULEVERSION = '1.0.0'

lang = {
    'exceptions' : {
        'runtimeheader'  : ' RUNTIME ERROR ',
        'runtimefooter'  : '',
        'syntaxheader'   : ' SYNTAX ERROR ',
        'syntaxfooter'   : '',
        'file'           : 'File',
        'module'         : 'Module',
        'in'             : 'In',
        'line'           : 'Line',
        'column'         : 'Column',
        'tracebackheader': 'Traceback (Most recent call last):',

        'argumenterror': {
            'name': 'ArgumentException'
        },
        'attributeerror': {
            'name': 'AttributeException'
        },
        'assertionerror': {
            'name': 'AssertionException'
        },
        'breakerror': {
            'name': 'BreakException',
            'location': 'Can not break from outside loop'
        },
        'continueerror': {
            'name': 'ContinueException',
            'location': 'Can not continue from outside loop'
        },
        'fileaccesserror': {
                'name': 'FileAccessException'
        },
        'identifiererror': {
            'name': 'IdentifierException',
            'notdefined': '\'%s\' is not defined'
        },
        'includeerror': {
            'name': 'IncludeException',
            'failed': 'Failed to include %s:\n  %s \'%s\', %s %s\n%s: %s',
            'doesnotexist': 'Module %s does not exist'
        },
            'indexerror': {
            'name': 'IndexException'
        },
        'iterationerror': {
            'name': 'IterationException',
            'cannot': '%s is not iterable'
        },
        'keyerror': {
            'name': 'KeyException'
        },
        'operationerror': {
            'name': 'OperationException'
        },
        'panicerror': {
            'name': 'PanicException'
        },
        'patternerror': {
            'name': 'PatternException',
            'nomatch': '%s does not match pattern %s'
        },
        'returnerror': {
            'name': 'ReturnException',
            'location': 'Can not return from outside function'
        },
        'reservederror': {
            'name': 'ReservedNameException',
            'reserved': 'Can not assign %s to \'%s\' (reserved)'
        },
        'typeerror': {
            'name': 'TypeException',
            'cannot': '%s of type %s can not %s %s',
            'assigntype': 'Can not assign %s to \'%s\' (%s)',
            'mustbe': '%s must be of type %s, %s given'
        },
        'valueerror': {
            'name': 'ValueException'
        },

        'syntaxerror': {
            'name': 'SyntaxException',
            'illegalchar': 'Illegal character \'%s\' was found',
            'invalideofl': 'Invalid EO%s, expected %s',
            'invalideoflnoexp': 'Invalid EO%s',
            'notfound': 'Expected %s not found'
        },
        'escapeerror': {
            'name': 'EscapeException',
            'cannot': '\'%s\' can not be escaped'
        }
    }
}

def main():
    from .version import catch as catch

    @catch.catch
    def improvederrormessage():
        ##########################################
        # DEPENDENCIES                           #
        ##########################################

        import sys
        from pathlib import Path

        import click
        from   colorama  import init, Fore, Style
        init()

        from .version import constants          as constants
        from .version import context            as context
        from .version import default            as default
        from .version import exceptions         as exceptions
        from .version import interpreter        as interpreter
        from .version import lexer              as lexer
        from .version import nodes              as nodes
        from .version import parser             as parser
        from .version import repl               as i_repl
        from .version import run                as run
        from .version import tokens             as tokens
        from .version import types              as types
        import perimod

        ##########################################
        # LOGO                                   #
        ##########################################

        def logo() -> str:
            logolines = [
                f'    _____----_____',
                f'   / \ PERI.DOT / \   Peri.dot - {VERSION}',
                f' _/  /\ __     /\  \_  © 2020 Totobird Creations',
                f'/___/  \__\-,  \ \___\ ',
                f'\‾‾‾\ \  \'-\‾‾\  /‾‾‾/ ',
                f' ‾\  \/     ‾‾ \/  /‾  https://github.com/toto-bird/Peri.dot',
                f'   \ / LANGUAGE \ /',
                f'    ‾‾‾‾‾----‾‾‾‾‾'
            ]
            logolength = 0
            for i in range(len(logolines)):
                j = logolines[i]
                if len(j) > logolength:
                    logolength = len(j)
            for i in range(len(logolines)):
                logolines[i] = f' ║ {logolines[i].ljust(logolength, " ")} ║'
            logolines.insert(0, f' ╔{"═" * (logolength + 2)}╗')
            logolines.append(f' ╚{"═" * (logolength + 2)}╝')
            logo = '\n'.join(logolines)
            logo = f'{Fore.GREEN}{Style.BRIGHT}{logo}{Style.RESET_ALL}'
            print(logo, end='\n')
            return(logo)

        ##########################################
        # ARGUMENTS                              #
        ##########################################

        @click.command()
        @click.option('-h', '--help',    is_flag=True)
        @click.option('-v', '--version', is_flag=True)
        @click.option('-r', '--repl',    is_flag=True)
        @click.argument('filename', default='')
        @click.argument('args', nargs=-1)
        def main(help, version, repl, filename, args):
            if version:
                logo()

            if help:
                print(f'''{Fore.YELLOW}Usage{Style.RESET_ALL}: {Fore.YELLOW}{Style.BRIGHT}{sys.argv[0]} [OPTIONS]* [FILE]? [ARGS]*{Style.RESET_ALL}

{Fore.BLUE}{Style.BRIGHT}Options:{Style.RESET_ALL}
  {Fore.GREEN}{Style.BRIGHT}-h{Style.RESET_ALL}, {Fore.GREEN}{Style.BRIGHT}--help{Style.RESET_ALL}    - {Fore.GREEN}Display this help message.{Style.RESET_ALL}
  {Fore.GREEN}{Style.BRIGHT}-v{Style.RESET_ALL}, {Fore.GREEN}{Style.BRIGHT}--version{Style.RESET_ALL} - {Fore.GREEN}Display logo and version.{Style.RESET_ALL}
  {Fore.GREEN}{Style.BRIGHT}-r{Style.RESET_ALL}, {Fore.GREEN}{Style.BRIGHT}--repl{Style.RESET_ALL}    - {Fore.GREEN}Enter the repl.{Style.RESET_ALL}''')


            path = str(Path(filename).parent.resolve())

            default._defaultinit(MODULEVERSION, path, types, context)
            exceptions._exceptionsinit(lang)
            interpreter._interpreterinit(lang, tokens, context, default, constants, types, exceptions, run, perimod)
            lexer._lexerinit(lang, constants, tokens, exceptions)
            parser._parserinit(lang, tokens, exceptions, constants, nodes)
            perimod._perimodinit(catch, types, interpreter, exceptions)
            i_repl._replinit(default, context, run)
            run._runinit(lexer, parser, context, interpreter)
            types._typesinit(catch, exceptions, context, constants, tokens, nodes, interpreter)

            if filename:
                script = ''

                try:
                    with open(filename, 'r') as f:
                        script = f.read()
                        script = '\n'.join(
                            [
                                i.lstrip(' ').rstrip(' ').lstrip('\t').rstrip('\t') for i in script.split('\n')
                            ]
                        )

                except Exception as e:
                    exc = sys.exc_info()
                    print(exceptions.Cmd_CmdArgumentError(f'{exc[0].__name__}: {str(e)}', 'filename', filename).asstring())
                    exit(1)

                symbols = default.defaultvariables(context.SymbolTable())
                result, exec_context, error = run.run(filename, script, symbols)

                if error:
                    print(error.asstring())
                    exit(1)


            if not (filename or version or help) or repl:
                #print(
                #    Cmd_OutOfDateWarning(f'\n  The REPL is currently out of date and does not have the same functionality as running a file.\n    Proceed with caution as it may crash or have bugs.\n\n    {Style.BRIGHT}PLEASE DO NOT REPORT ANY INTERNAL ERRORS CAUSED BY THE REPL ON GITHUB').asstring()
                #)

                i_repl.Repl()


        main()

if __name__ == "__main__":
    main()



##########################################
# DEPENDENCIES                           #
##########################################

def _defaultinit(appversion, file, types, context):
    global  __file__, version,SymbolTable
    __file__            = file
    version             = appversion
    SymbolTable         = context.SymbolTable
    global TYPES, ArrayType, BooleanType, BuiltInFunctionType, FloatType, IdType, IntType, NamespaceType, NullType, StringType, TupleType
    TYPES               = types.TYPES
    ArrayType           = types.ArrayType
    BooleanType         = types.BooleanType
    BuiltInFunctionType = types.BuiltInFunctionType
    FloatType           = types.FloatType
    IdType              = types.IdType
    IntType             = types.IntType
    NamespaceType       = types.NamespaceType
    NullType            = types.NullType
    StringType          = types.StringType
    TupleType           = types.TupleType

##########################################
# DEFAULT VARIABLES                      #
##########################################

import xdgappdirs as _xdgappdirs
from pathlib import Path as _Path

def defaultvariables(symbols):
    perisymbols = SymbolTable(symbols)
    perisymbols.assign('path', ArrayType([
        StringType(__file__),
        StringType(str(_xdgappdirs.user_data_dir('Peridot', 'TotobirdCreations', as_path=True) / 'modules' / version)),
        StringType(str(_xdgappdirs.site_data_dir('Peridot', 'TotobirdCreations', as_path=True) / 'modules' / version))
    ]))
    symbols.assign('__peridot__', NamespaceType(perisymbols))

    symbols.assign('True'  , BooleanType(True))
    symbols.assign('False' , BooleanType(False))
    symbols.assign('Null'  , NullType())

    symbols.assign('throw' , BuiltInFunctionType('throw'))
    symbols.assign('assert', BuiltInFunctionType('assert'))
    symbols.assign('panic' , BuiltInFunctionType('panic'))
    symbols.assign('print' , BuiltInFunctionType('print'))
    symbols.assign('range' , BuiltInFunctionType('range'))

    symbols.assign('type'  , BuiltInFunctionType('type', type_=TYPES['type'], returntype=BuiltInFunctionType))
    symbols.assign('id'    , BuiltInFunctionType('id', type_=TYPES['type'], returntype=IdType))
    symbols.assign('str'   , BuiltInFunctionType('str', type_=TYPES['type'], returntype=StringType))
    symbols.assign('int'   , BuiltInFunctionType('int', type_=TYPES['type'], returntype=IntType))
    symbols.assign('float' , BuiltInFunctionType('float', type_=TYPES['type'], returntype=FloatType))
    symbols.assign('bool'  , BuiltInFunctionType('bool', type_=TYPES['type'], returntype=BooleanType))
    symbols.assign('array' , BuiltInFunctionType('array', type_=TYPES['type'], returntype=ArrayType))
    symbols.assign('tuple' , BuiltInFunctionType('tuple', type_=TYPES['type'], returntype=TupleType))

    return(symbols)


import os as _os
from traceback import format_exc as _format_exc
from colorama import init as _init, Fore as _Fore, Style as _Style
_init()

class InternalPeridotError(BaseException): pass

_HEADER = '\n- INTERNAL ERROR '
_FOOTER = f'{_Fore.BLUE}{_Style.BRIGHT}Please report this error on github\n  (https://github.com/toto-bird/Peri.dot/issues/new){_Style.RESET_ALL}'

def catch(func):
    try:
        func()



    except Exception as e:
        size = _os.get_terminal_size()
        traceback = _format_exc()
        print(f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{_HEADER}{"-" * max([size[0] - len(_HEADER) + 1, 0])}{_Style.RESET_ALL}\n')

        traceback = traceback.rstrip('\n').split('\n\n')
        traceback = [i.split('\n') for i in traceback]

        for j in range(len(traceback)):
            lines = traceback[j]

            for i in range(len(lines)):
                ln = lines[i]


                if ln == 'During handling of the above exception, another exception occurred:':
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif i == len(lines) - 1:
                    ln = ln.split(' ')
                    ln = [ln[0][:-1], ' '.join(ln[1:])]
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln[0]}{_Style.RESET_ALL}: {_Fore.RED}{ln[1]}{_Style.RESET_ALL}')

                elif ln == 'Traceback (most recent call last):':
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif ln.startswith('    '):
                    ln = ln.lstrip(' ')
                    print(f'      {_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif ln.startswith('  '):
                    ln = ln.lstrip(' ').split(' ')
                    ln = [ln[1][1:-2], ln[-3][:-1], ln[-1]]
                    print(f'  {_Fore.RED}File {_Style.BRIGHT}\'{ln[0]}\'{_Style.RESET_ALL}')
                    print(f'    {_Fore.RED}Line {_Style.BRIGHT}{ln[1]}{_Style.RESET_ALL} {_Fore.RED}In {_Style.BRIGHT}{ln[2]}{_Style.RESET_ALL}')

                else:
                    print(ln)


            if j < len(traceback) - 1:
                print('')


        print(f'\n{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{"-" * size[0]}{_Style.RESET_ALL}')
        print(f'{_FOOTER}')
        print(f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{"-" * size[0]}{_Style.RESET_ALL}')



    except InternalPeridotError as e:
        size = _os.get_terminal_size()
        traceback = _format_exc()
        print(f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{_HEADER}{"-" * max([size[0] - len(_HEADER) + 1, 0])}{_Style.RESET_ALL}')

        traceback = traceback.rstrip('\n').split('\n\n')
        traceback = [i.split('\n') for i in traceback]

        for j in range(len(traceback)):
            lines = traceback[j]

            for i in range(len(lines)):
                ln = lines[i]


                if ln == 'During handling of the above exception, another exception occurred:':
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif i == len(lines) - 1:
                    ln = ln.split(' ')
                    ln = [ln[0][:-1], ' '.join(ln[1:])]
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln[0]}{_Style.RESET_ALL}: {_Fore.RED}{ln[1]}{_Style.RESET_ALL}')

                elif ln == 'Traceback (most recent call last):':
                    print(f'{_Fore.RED}{_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif ln.startswith('    '):
                    ln = ln.lstrip(' ')
                    print(f'      {_Style.BRIGHT}{ln}{_Style.RESET_ALL}')

                elif ln.startswith('  '):
                    ln = ln.lstrip(' ').split(' ')
                    ln = [ln[1][1:-2], ln[-3][:-1], ln[-1]]
                    print(f'  {_Fore.RED}File {_Style.BRIGHT}\'{ln[0]}\'{_Style.RESET_ALL}')
                    print(f'    {_Fore.RED}Line {_Style.BRIGHT}{ln[1]}{_Style.RESET_ALL} {_Fore.RED}In {_Style.BRIGHT}{ln[2]}{_Style.RESET_ALL}')

                else:
                    print(ln)


            if j < len(traceback) - 1:
                print('')


        print(f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{"-" * size[0]}{_Style.RESET_ALL}')
        print(f'{_FOOTER}')
        print(f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{"-" * size[0]}{_Style.RESET_ALL}')



##########################################
# DEPENDENCIES                           #
##########################################

def nodesinit(types, tokens):
    global TYPES
    TYPES      = types.TYPES
    global TT_KEYWORD
    TT_KEYWORD = tokens.TT_KEYWORD
    

##########################################
# NODES                                  #
##########################################

### TYPES
class IntNode():
    def __init__(self, token):
        self.token = token

        self.start = self.token.start
        self.end   = self.token.end

    def __repr__(self):
        return(f'{self.token.value}')


class FloatNode():
    def __init__(self, token):
        self.token = token

        self.start = self.token.start
        self.end   = self.token.end

    def __repr__(self):
        return(f'{self.token.value}')


class StringNode():
    def __init__(self, token):
        self.token = token

        self.start = self.token.start
        self.end   = self.token.end

    def __repr__(self):
        return(f'\'{self.token.value}\'')


class ArrayNode():
    def __init__(self, elmnodes, start, end):
        self.elmnodes = elmnodes

        self.start = start
        self.end   = end

    def __repr__(self):
        return(f'[{", ".join([str(i) for i in self.elmnodes])}]')


class DictionaryNode():
    def __init__(self, keynodes, valuenodes, start, end):
        self.keynodes = keynodes
        self.valuenodes = valuenodes

        self.start = start
        self.end   = end

    def __repr__(self):
        result = ''
        first = True
        keys = self.keynodes
        values = self.valuenodes
        for i in range(len(keys)):
            if not first:
                result += ', '
            result += f'{keys[i]}: {values[i]}'
            first = False
        return(f'{{{result}}}')


class TupleNode():
    def __init__(self, elmnodes, start, end):
        self.elmnodes = elmnodes

        self.start = start
        self.end   = end

    def __repr__(self):
        return(f'({", ".join([str(i) for i in self.elmnodes])})')



### VARIABLE CONTROL
class VarAccessNode():
    def __init__(self, token):
        self.token = token

        self.start = self.token.start
        self.end   = self.token.end

    def __repr__(self):
        return(f'{self.token.value}')


class VarAssignNode():
    def __init__(self, token, valnode):
        self.token = token
        self.valnode = valnode

        self.start = token.start
        self.end = token.end

    def __repr__(self):
        return(f'{self.valnode}')


class VarCreateNode():
    def __init__(self, token, valnode):
        self.token = token
        self.valnode = valnode

        self.start = token.start
        self.end = token.end

    def __repr__(self):
        return(f'{self.valnode}')


class VarNullNode():
    def __init__(self, token):
        self.token = token

        self.start = token.start
        self.end = token.end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')



### FUNCTIONS
class FuncCreateNode():
    def __init__(self, token, arguments, options, returntype, bodynodes, shouldreturn, end=None):
        self.token = token
        self.arguments = arguments
        self.options = options
        self.returntype = returntype
        self.bodynodes = bodynodes
        self.shouldreturn = shouldreturn

        self.start = self.token.start
        if end:
            self.end = end
        else:
            self.end = self.token.end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class FuncCallNode():
    def __init__(self, node, calls):
        self.node = node
        self.calls = calls

        self.start = node.start
        self.end = calls[-1][2]

    def __repr__(self):
        return(f'Return value of {self.node}')


class ReturnNode():
    def __init__(self, returnnode, start, end):
        self.returnnode = returnnode

        self.start = start
        self.end = end

    def __repr__(self):
        return(f'{self.returnnode}')



### HANDLER
class HandlerNode():
    def __init__(self, token, bodynodes):
        self.token = token
        self.bodynodes = bodynodes

        self.start = self.token.start
        self.end = self.token.end

        if len(self.bodynodes) > 0:
            self.end = self.bodynodes[-1].end

    def __repr__(self):
        return(f'Result of handler')


### FLOW CONTROL
class IfNode():
    def __init__(self, cases, elsecase):
        self.cases = cases
        self.elsecase = elsecase

        self.start = self.cases[0][0].start
        if self.elsecase:
            self.end = self.elsecase[-1].end
        else:
            self.end = self.cases[-1][0].end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class SwitchNode():
    def __init__(self, vartoken, varoverwrite, value, cases, elsecase, start, end):
        self.vartoken = vartoken
        self.varoverwrite = varoverwrite
        self.value = value
        self.cases = cases
        self.elsecase = elsecase

        self.start = start
        self.end = end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class ForLoopNode():
    def __init__(self, vartoken, varoverwrite, loopthrough, bodynodes):
        self.vartoken = vartoken
        self.varoverwrite = varoverwrite
        self.loopthrough = loopthrough
        self.bodynodes = bodynodes

        self.start = self.vartoken.start

        if len(self.bodynodes) > 0:
            self.end = self.bodynodes[-1].end
        else:
            self.end = self.vartoken.end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class WhileLoopNode():
    def __init__(self, condition, bodynodes):
        self.condition = condition
        self.bodynodes = bodynodes

        self.start = self.condition.start

        if len(self.bodynodes) > 0:
            self.end = self.bodynodes[-1].end
        else:
            self.end = self.condition.end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class BreakNode():
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')


class ContinueNode():
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return(f'{TYPES["nonetype"]}')



### OPERATIONS
class UnaryOpNode():
    def __init__(self, optoken, node):
        self.optoken = optoken
        self.node   = node

        self.start = self.optoken.start
        self.end   = self.node.end

    def __repr__(self):
        return(f'({self.optoken.type} {self.node})')


class BinaryOpNode():
    def __init__(self, lnode, optoken, rnode):
        self.lnode   = lnode
        self.optoken = optoken
        self.rnode   = rnode

        self.start = self.lnode.start
        self.end   = self.rnode.end

    def __repr__(self):
        if self.optoken.type == TT_KEYWORD:
            return(f'({self.lnode} {self.optoken.value} {self.rnode})')
        else:
            return(f'({self.lnode} {self.optoken.type} {self.rnode})')



### MISCELLANIOUS
class IndicieNode():
    def __init__(self, node, indicies, end=None):
        self.node = node
        self.indicies = indicies
        
        self.start = self.node.start
        if end:
            self.end = end
        else:
            self.end = self.node.end

    def __repr__(self):
        indicies = ''
        for i in self.indicies:
            indicies += f'[{i}]'
        return(f'{self.node}{indicies}')


class AttributeNode():
    def __init__(self, node, attributes, end=None):
        self.node = node
        self.attributes = attributes

        self.start = self.node.start
        if end:
            self.end = end
        else:
            self.end = self.node.end

    def __repr__(self):
        attributes = ''
        for i in self.attributes:
            attributes += f'.{i.value}'
        return(f'{self.node}{attributes}')



### MODULES
class IncludeNode():
    def __init__(self, filenode, start, end):
        self.filenode = filenode

        self.start = start
        self.end = end

    def __repr__(self):
        return(f'Namespace {self.filenode}')



##########################################
# DEPENDENCIES                           #
##########################################

def _replinit(default, context, runu):
    global defaultvariables
    defaultvariables = default.defaultvariables
    global SymbolTable
    SymbolTable      = context.SymbolTable
    global run
    run              = runu.run


##########################################
# REPL                                   #
##########################################

replname = '<repl>'

class Repl():
    def __init__(self):
        end = False
        symbols = defaultvariables(SymbolTable())
        failed = False

        while not end:
            script = input('>>> ')

            result, context, error = run(replname, script, symbols)

            if error:
                print(error.asstring())

            if result:
                print('\n'.join(str(i) for i in result))

            print('')



#class CursesRepl():
#    def __init__(self):
#        curses.wrapper(self.main)
#
#    def main(self, stdscr):
#        if curses.can_change_color():
#            curses.init_color(0, 0, 0, 0)
#
#        curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
#        curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_BLACK)
#
#        stdscr.clear()
#
#        while True:
#            try:
#                size = stdscr.getmaxyx()
#                if size[0] >= 2:
#                    stdscr.addstr(
#                        0, 0,
#                        TITLE.center(size[1], ' '),
#                        curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD
#                    )
#                if size[0] >= 1:
#                    stdscr.insstr(
#                        size[0] - 1, 0,
#                        TITLE.center(size[1], ' '),
#                        curses.color_pair(2) | curses.A_REVERSE
#                    )
#
#                stdscr.refresh()
#
#            except KeyboardInterrupt as e:
#                raise KeyboardInterrupt(str(e))
